'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _asyncToGenerator2 = require('babel-runtime/helpers/asyncToGenerator');

var _asyncToGenerator3 = _interopRequireDefault(_asyncToGenerator2);

var _getPrototypeOf = require('babel-runtime/core-js/object/get-prototype-of');

var _getPrototypeOf2 = _interopRequireDefault(_getPrototypeOf);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

var _possibleConstructorReturn2 = require('babel-runtime/helpers/possibleConstructorReturn');

var _possibleConstructorReturn3 = _interopRequireDefault(_possibleConstructorReturn2);

var _inherits2 = require('babel-runtime/helpers/inherits');

var _inherits3 = _interopRequireDefault(_inherits2);

var _lodash = require('lodash');

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _defaultOptions = require('./default-options');

var _defaultOptions2 = _interopRequireDefault(_defaultOptions);

var _base = require('../base');

var _base2 = _interopRequireDefault(_base);

var _fsExtraPromisify = require('fs-extra-promisify');

var _fsExtraPromisify2 = _interopRequireDefault(_fsExtraPromisify);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/// @name Folder
/// @page api
/// @description This is used to output data to the Folder
var Folder = function (_Base) {
  (0, _inherits3.default)(Folder, _Base);

  ///# @name constructor
  ///# @arg {object} options - The base options
  ///# @arg {object} output_options - The output options
  function Folder() {
    var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var output_options = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    (0, _classCallCheck3.default)(this, Folder);

    var _this = (0, _possibleConstructorReturn3.default)(this, (Folder.__proto__ || (0, _getPrototypeOf2.default)(Folder)).call(this, options));

    _this.output_options = (0, _lodash.extend)({}, _defaultOptions2.default, output_options);

    _this.prepared = false;
    return _this;
  }

  ///# @name prepare
  ///# @description
  ///# This is used to prepare the saving functionality that is determined by the
  ///# options that were passed to the constructor.
  ///# It sets a variable of `this.preparing` that ultimately calls `this.setup` that returns a promise.
  ///# This way when you go to save data it, that function will know if the setup is complete or not and
  ///# wait for it to be done before it starts saving data.
  ///# @returns {promise} - The setup function that was called
  ///# @async


  (0, _createClass3.default)(Folder, [{
    key: 'prepare',
    value: function prepare() {
      this.preparing = true;
      this.preparing = this.setup();
      return this.preparing;
    }

    ///# @name setup
    ///# @description
    ///# This is used to setup the saving function that will be used.

  }, {
    key: 'setup',
    value: function () {
      var _ref = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee() {
        return _regenerator2.default.wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (!(this.preparing == null)) {
                  _context.next = 2;
                  break;
                }

                return _context.abrupt('return', this.prepare());

              case 2:

                // Resolve the output path to make sure it's absolute
                this.output_options.output = this.resolvePaths(this.output_options.output)[0];

                // Ensure there is a directory to write to. This way we can use `fs.writeFile`
                // instead of `fs.outputFile` which has extra checks that we can skip over
                _context.next = 5;
                return _fsExtraPromisify2.default.ensureDir(this.output_options.output);

              case 5:

                this.prepared = true;

              case 6:
              case 'end':
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function setup() {
        return _ref.apply(this, arguments);
      }

      return setup;
    }()

    ///# @name output
    ///# @description
    ///# This is used to output the data that's passed to it
    ///# @arg {string} id - The id to use for this data
    ///# @arg {object, array, string} data - The data that you want to be saved
    ///# @async

  }, {
    key: 'output',
    value: function () {
      var _ref2 = (0, _asyncToGenerator3.default)( /*#__PURE__*/_regenerator2.default.mark(function _callee2(id, data) {
        return _regenerator2.default.wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (!(this.prepared !== true)) {
                  _context2.next = 4;
                  break;
                }

                if (this.preparing == null) {
                  this.prepare();
                }
                _context2.next = 4;
                return this.preparing;

              case 4:
                return _context2.abrupt('return', _fsExtraPromisify2.default.writeFile(_path2.default.join(this.output_options.output, id + '.' + this.output_options.format), data));

              case 5:
              case 'end':
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function output(_x3, _x4) {
        return _ref2.apply(this, arguments);
      }

      return output;
    }()
  }]);
  return Folder;
}(_base2.default);

exports.default = Folder;